//   Copyright (c)  2005  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License (version 2)
//   as published by the Free Software Foundation.  A copy of the full
//   licence may be found in the file COPYING in this directory.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoA; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


// Source code for classes DenseMatrix and DenseMatrixImpl

#include "CoCoA/DenseMatrix.H"
#include "CoCoA/FractionField.H"
#include "CoCoA/MatrixArith.H"
#include "CoCoA/assert.H"
#include "CoCoA/error.H"
#include "CoCoA/matrix.H"
#include "CoCoA/ring.H"


//#include <cstddef>
using std::size_t;
//#include <vector>
using std::vector;
#include <algorithm>
//using std::swap;


namespace CoCoA
{

  class DenseMatrixBase: public MatrixBase
  {
  protected: // data members
    const ring myR;
    std::size_t myNumRowsValue;
    std::size_t myNumColsValue;
  protected:
    DenseMatrixBase(const ring& R, std::size_t nrows, std::size_t ncols);
  };

//   class DenseMatrix;

  class DenseMatrixImpl: public DenseMatrixBase
  {
  public:
    DenseMatrixImpl(const ring& R, std::size_t nrows, std::size_t ncols); // creates zero matrix
    ~DenseMatrixImpl();
  private: // implementation detail
    void myDtorBody();
  public:
//???    MatrixBase* clone() const = 0;
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
//???    virtual RingRawValue& myRawEntry(std::size_t i, std::size_t j);
//???    virtual const RingRawValue& myRawEntry(std::size_t i, std::size_t j) const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
    virtual DenseMatrixImpl* myZeroClone(const ring& R, std::size_t NumRows, std::size_t NumCols) const;
    virtual void mySwap(matrix& M); // homogeneous swap
    virtual void myAssignZero(); // sets all entries to zero
    virtual void mySetEntry(std::size_t i, std::size_t j, ConstRefRingElem r);
    virtual void mySetEntry(std::size_t i, std::size_t j, long n);
    virtual void mySetEntry(std::size_t i, std::size_t j, const ZZ& N);
// Use generic versions of the following four functions.
//     virtual void myMulByRow(vec& lhs, const vec& v) const;
//     virtual void myMulByCol(vec& lhs, const vec& v) const;
//     virtual bool myIsZeroRow(std::size_t i) const; ///< tests whether row i is zero
//     virtual bool myIsZeroCol(std::size_t j) const; ///< tests whether col j is zero
    virtual void myRowMul(std::size_t i, ConstRefRingElem c); ///< row(i) = c*row(i)
    virtual void myColMul(std::size_t j, ConstRefRingElem c); ///< row(j) = c*row(j)
    virtual void myAddRowMul(std::size_t i1, std::size_t i2, ConstRefRingElem c); ///< row(i1) += c*row(i2)
    virtual void myAddColMul(std::size_t j1, std::size_t j2, ConstRefRingElem c); ///< row(j1) += c*row(j2)
    virtual void mySwapRows(std::size_t i1, std::size_t i2);
    virtual void mySwapCols(std::size_t j1, std::size_t j2);
    virtual void myDet(RefRingElem d) const;

    friend void inverse(matrix& lhs, ConstMatrix M); // lhs=inv(M)
    friend void adjoint(matrix& lhs, ConstMatrix M); // lhs=adj(M)
    friend void mul(matrix& lhs, ConstMatrix M1, ConstMatrix M2); // lhs = M1*M2
    virtual size_t myRank() const;
   private: // data members
    std::vector< std::vector< RingElemRawPtr > > myEntries;
  };



//  BETTER NOT TO HAVE A SEPARATE public smart pointer class
//   class DenseMatrix: public matrix
//   {
//     // smart ref counted pointer class???
//   public:
//     DenseMatrix(DenseMatrixImpl* MatPtr); // public???
//     DenseMatrix(const DenseMatrix& M);
//     DenseMatrix& operator=(const DenseMatrix&);
//     ~DenseMatrix();
//     DenseMatrixImpl* operator->();             // covariant return type
//     const DenseMatrixImpl* operator->() const; // covariant return type
//   };


  /////////////////////////////////////////////////////////////////////////////
  // Inline functions

  inline DenseMatrixBase::DenseMatrixBase(const ring& R, std::size_t NumRows, std::size_t NumCols):
      myR(R),
      myNumRowsValue(NumRows),
      myNumColsValue(NumCols)
  {}


//   inline DenseMatrix::DenseMatrix(DenseMatrixImpl* MatPtr):
//       matrix(MatPtr)
//   {}


//   inline DenseMatrix::~DenseMatrix()
//   {} // dtor for matrix does all the work (I hope ???!!!???)


//   inline DenseMatrixImpl* DenseMatrix::operator->()
//   {
//     return static_cast<DenseMatrixImpl*>(myMatrixPtr());
//   }

//   inline const DenseMatrixImpl* DenseMatrix::operator->() const
//   {
//     return static_cast<const DenseMatrixImpl*>(myMatrixPtr());
//   }


//----------------------------------------------------------------------

  DenseMatrixImpl::DenseMatrixImpl(const ring& R, size_t NumRows, size_t NumCols):
      DenseMatrixBase(R, NumRows, NumCols)
  {
    // Making this exception safe turns out to be a bit complicated.
    try
    {
      myEntries.resize(myNumRows()); // not necessary but might help
      size_t i, j;
      for (i=0; i < myNumRows(); ++i)
      {
        myEntries.push_back(vector<RingElemRawPtr>());
        myEntries[i].reserve(myNumCols()); // not necessary but might help
        for (j=0; j < myNumCols(); ++j)
        {
          myEntries[i].push_back(myR->myNew());
        }
      }
    }
    catch (...)
    {
      myDtorBody();
    }
  }


  DenseMatrixImpl::~DenseMatrixImpl()
  {
    myDtorBody();
  }


  // Ignore the values of myNumRows and myNumCols because this procedure
  // may be called on a partially constructed DenseMatrix if an exception
  // is thrown during the ctor.
  void DenseMatrixImpl::myDtorBody()
  {
    size_t i, j;
    // These two loops are slightly funny because i,j are unsigned.
    // Written this way, we destroy in reverse order of construction.
    for (i=myEntries.size(); i > 0; --i)
    {
      for (j=myEntries[i-1].size(); j > 0; --j)
        myR->myDelete(myEntries[i-1][j-1]);
    }
  }


  const ring& DenseMatrixImpl::myBaseRing() const
  {
    return myR;
  }


  size_t DenseMatrixImpl::myNumRows() const
  {
    return myNumRowsValue;
  }


  size_t DenseMatrixImpl::myNumCols() const
  {
    return myNumColsValue;
  }


//   RingRawValue& DenseMatrixImpl::myRawEntry(size_t i, size_t j)
//   {
//     CoCoA_ASSERT(i < myNumRows());
//     CoCoA_ASSERT(j < myNumCols());
//     return raw(myEntries[i][j]);
//   }


//   const RingRawValue& DenseMatrixImpl::myRawEntry(size_t i, size_t j) const
//   {
//     CoCoA_ASSERT(i < myNumRows());
//     CoCoA_ASSERT(j < myNumCols());
//     return raw(myEntries[i][j]);
//   }


  inline ConstRefRingElem DenseMatrixImpl::myEntry(size_t i, size_t j) const
  {
    CoCoA_ASSERT(i < myNumRows());
    CoCoA_ASSERT(j < myNumCols());
    return ConstRefRingElem(myR, myEntries[i][j]);
  }


  DenseMatrixImpl* DenseMatrixImpl::myZeroClone(const ring& R, std::size_t NumRows, std::size_t NumCols) const
  {
    return new DenseMatrixImpl(R, NumRows, NumCols);
  }


  void DenseMatrixImpl::mySwap(matrix& M)
  {
    DenseMatrixImpl* other = dynamic_cast<DenseMatrixImpl*>(M.myMatrixPtr());
    CoCoA_ASSERT(other != 0);
    CoCoA_ASSERT(myBaseRing() == BaseRing(M));
    std::swap(myEntries, other->myEntries);
    std::swap(myNumRowsValue, other->myNumRowsValue);
    std::swap(myNumColsValue, other->myNumColsValue);
  }


  void DenseMatrixImpl::myAssignZero()
  {
    // This fn is probably not exception safe -- can myAssignZero throw?
    for (size_t i=0; i < myNumRows(); ++i)
      for (size_t j=0; j < myNumCols(); ++j)
        myR->myAssignZero(myEntries[i][j]);
  }


  void DenseMatrixImpl::mySetEntry(size_t i, size_t j, ConstRefRingElem r)
  {
    CoCoA_ASSERT(owner(r) == myR);
    CoCoA_ASSERT(i < myNumRows());
    CoCoA_ASSERT(j < myNumCols());
    myR->myAssign(myEntries[i][j], raw(r));
  }


  void DenseMatrixImpl::mySetEntry(size_t i, size_t j, long n)
  {
    CoCoA_ASSERT(i < myNumRows());
    CoCoA_ASSERT(j < myNumCols());
    myR->myAssign(myEntries[i][j], n);
  }


  void DenseMatrixImpl::mySetEntry(size_t i, size_t j, const ZZ& N)
  {
    CoCoA_ASSERT(i < myNumRows());
    CoCoA_ASSERT(j < myNumCols());
    myR->myAssign(myEntries[i][j], N);
  }


//   MOVED UP TO GENERIC IMPLEMENTATIONS IN ConstMatrixBase
//   void DenseMatrixImpl::myMulByRow(vec& lhs, const vec& v) const
//   {
//     CoCoA_ASSERT(lhs.size() == myNumCols() && v.size() == myNumRows());
//     vector<RingElem> ans;
//     ans.resize(myNumCols(), zero(myR));
//     for (size_t j=0; j < myNumCols(); ++j)
//       for (size_t i=0; i < myNumRows(); ++i)
//         ans[j] += v[i]*myEntry(i, j);
//     // We have successfully computed the answer, now swap it in.
//     swap(lhs, ans);
// //     for (size_t j=0; j < myNumCols(); ++j)
// //       swap(lhs[j], ans[j]);
//   }


//   void DenseMatrixImpl::myMulByCol(vec& lhs, const vec& v) const
//   {
//     CoCoA_ASSERT(lhs.size() == myNumRows() && v.size() == myNumCols());
//     vector<RingElem> ans;
//     ans.resize(myNumRows(), zero(myR));
//     for (size_t i=0; i < myNumRows(); ++i)
//       for (size_t j=0; j < myNumCols(); ++j)
//         ans[i] += myEntry(i, j)*v[j];

//     // We have successfully computed the answer, now swap it in.
//     swap(lhs, ans);
// //     for (size_t i=0; i < myNumRows(); ++i)
// //       swap(lhs[i], ans[i]);
//   }


//   bool DenseMatrixImpl::myIsZeroRow(std::size_t i) const
//   {
//     if (i >= myNumRows())
//       CoCoA_ERROR(ERR::BadRowIndex, "DenseMatrixImpl::myIsZeroRow");
//     for (size_t j=0; j < myNumCols(); ++j)
//       if (!myR->myIsZero(myEntries[i][j])) return false;
//     return true;
//   }


//   bool DenseMatrixImpl::myIsZeroCol(std::size_t j) const
//   {
//     if (j >= myNumCols())
//       CoCoA_ERROR(ERR::BadColIndex, "DenseMatrixImpl::myIsZeroCol");
//     for (size_t i=0; i < myNumRows(); ++i)
//       if (!myR->myIsZero(myEntries[i][j])) return false;
//     return true;
//   }


  void DenseMatrixImpl::myRowMul(std::size_t i, ConstRefRingElem c)
  {
    if (i >= myNumRows())
      CoCoA_ERROR(ERR::BadRowIndex, "DenseMatrixImpl::myAddRowMul");
    if (owner(c) != myR)
      CoCoA_ERROR(ERR::MixedRings, "DenseMatrixImpl::myAddRowMul");
    if (IsOne(c)) return;
    vector<RingElem> ans;
    ans.resize(myNumCols(), zero(myR));
    for (size_t j = 0; j < myNumCols(); ++j)
      myR->myMul(raw(ans[j]), raw(c), myEntries[i][j]);
    // Answer successfully computed in ans, swap it into the i-th row.
    for (size_t j = 0; j < myNumCols(); ++j)
      myR->mySwap(myEntries[i][j], raw(ans[j]));
  }


  void DenseMatrixImpl::myColMul(std::size_t j, ConstRefRingElem c)
  {
    if (j >= myNumCols())
      CoCoA_ERROR(ERR::BadColIndex, "DenseMatrixImpl::myAddColMul");
    if (owner(c) != myR)
      CoCoA_ERROR(ERR::MixedRings, "DenseMatrixImpl::myAddColMul");
    if (IsOne(c)) return;
    vector<RingElem> ans;
    ans.resize(myNumRows(), zero(myR));
    for (size_t i = 0; i < myNumRows(); ++i)
      myR->myMul(raw(ans[i]), raw(c), myEntries[i][j]);
    // Answer successfully computed in ans, swap it into the i-th row.
    for (size_t i = 0; i < myNumRows(); ++i)
      myR->mySwap(myEntries[i][j], raw(ans[i]));
  }


  void DenseMatrixImpl::myAddRowMul(std::size_t i1, std::size_t i2, ConstRefRingElem c)
  {
    if (i1 >= myNumRows() || i2 >= myNumRows())
      CoCoA_ERROR(ERR::BadRowIndex, "DenseMatrixImpl::myAddRowMul");
    if (owner(c) != myR)
      CoCoA_ERROR(ERR::MixedRings, "DenseMatrixImpl::myAddRowMul");
    if (IsZero(c)) return;
    vector<RingElem> ans;
    ans.resize(myNumCols(), zero(myR));
    for (size_t j = 0; j < myNumCols(); ++j)
    {
      myR->myAssign(raw(ans[j]),myEntries[i1][j]);
      myR->myIsZeroAddMul(raw(ans[j]), raw(c), myEntries[i2][j]);
    }
    // Answer successfully computed in ans, swap it into the i-th row
    for (size_t j = 0; j < myNumCols(); ++j)
      myR->mySwap(raw(ans[j]), myEntries[i1][j]);
  }


  void DenseMatrixImpl::myAddColMul(std::size_t j1, std::size_t j2, ConstRefRingElem c)
  {
    if (j1 >= myNumCols() || j2 >= myNumCols())
      CoCoA_ERROR(ERR::BadColIndex, "DenseMatrixImpl::myAddColMul");
    if (owner(c) != myR)
      CoCoA_ERROR(ERR::MixedRings, "DenseMatrixImpl::myAddColMul");
    if (IsZero(c)) return;
    vector<RingElem> ans;
    ans.resize(myNumRows(), zero(myR));
    for (size_t i = 0; i < myNumRows(); ++i)
    {
      myR->myAssign(raw(ans[i]),myEntries[i][j1]);
      myR->myIsZeroAddMul(raw(ans[i]), raw(c), myEntries[i][j2]);
    }
    // Answer successfully computed in ans, swap it into the i-th row
    for (size_t i = 0; i < myNumCols(); ++i)
      myR->mySwap(raw(ans[i]), myEntries[i][j1]);
  }


  void DenseMatrixImpl::mySwapRows(std::size_t i1, std::size_t i2)
  {
    if (i1 >= myNumRows() || i2 >= myNumRows())
      CoCoA_ERROR(ERR::BadRowIndex, "DenseMatrixImpl::mySwapRows");
    if (i1 == i2) return;
    for (size_t j=0; j < myNumCols(); ++j)
      myR->mySwap(myEntries[i1][j], myEntries[i2][j]);
  }


  void DenseMatrixImpl::mySwapCols(std::size_t j1, std::size_t j2)
  {
    if (j1 >= myNumCols() || j2 >= myNumCols())
      CoCoA_ERROR(ERR::BadColIndex, "DenseMatrixImpl::mySwapCols");
    if (j1 == j2) return;
    for (size_t i=0; i < myNumRows(); ++i)
      myR->mySwap(myEntries[i][j1], myEntries[i][j2]);
  }


  void DenseMatrixImpl::myDet(RefRingElem d) const
  {
    // ??? this function works only within integral domains!
    DetByGauss(d, ConstMatrix(this));
  }


  size_t DenseMatrixImpl::myRank() const
  {
    return RankByGauss(ConstMatrix(this));
  }


  void inverse(matrix& lhs, ConstMatrix M)
  {
    InverseByGauss(lhs, M);
  }


  void mul(matrix& lhs, ConstMatrix M1, ConstMatrix M2)
  {
    if (NumCols(M1) != NumRows(M2))
      CoCoA_ERROR(ERR::BadMatrixSize, "mul(matrix, matrix)");
    if ( NumRows(lhs) != NumRows(M1) || NumCols(lhs) != NumCols(M2) )
      CoCoA_ERROR(ERR::BadMatrixSize, "mul(matrix, matrix)");
    if (BaseRing(lhs) != BaseRing(M1) || BaseRing(lhs) != BaseRing(M2))
      CoCoA_ERROR(ERR::MixedRings, "mul(matrix, matrix)");
    // Use of the temporary ans avoids aliasing problems and makes the code exception safe.
    matrix ans(lhs->myZeroClone(BaseRing(lhs), NumRows(M1), NumCols(M2)));
    for (size_t i=0; i < NumRows(M1); ++i)
      for (size_t j=0; j < NumCols(M2); ++j)
        for (size_t k=0; k < NumCols(M1); ++k)
          SetEntry(ans, i, j, ans(i,j) + M1(i,k) * M2(k,j));

    // The answer is in ans; now swap it with the entries of lhs.
    lhs->mySwap(ans);
  }


  void adjoint(matrix& lhs, ConstMatrix M)
  {
    if (NumRows(M) != NumCols(M))
      CoCoA_ERROR(ERR::NotSquareMatrix, "adjoint(lhs, M)");
    // This code works only if the matrix is invertible!
    // It is morally equivalent to swap(lhs, inverse(M)*det(M));
    matrix tmp(lhs->myZeroClone(BaseRing(M), NumRows(M), NumCols(M)));
    inverse(tmp, M);
    RingElem d(BaseRing(M));
    M->myDet(d);
    for (size_t i=0; i < NumRows(M); ++i)
      for (size_t j=0; j < NumCols(M); ++j)
        SetEntry(tmp, i, j, tmp(i,j) * d);
    lhs->mySwap(tmp);
  }


  matrix NewDenseMatrix(const ring& R, size_t NumRows, size_t NumCols)
  {
    return matrix(new DenseMatrixImpl(R, NumRows, NumCols));
  }

  matrix NewDenseMatrix(const ring& R, std::size_t NumRows, std::size_t NumCols, const std::vector< std::vector<int> >& M)
  {
    // Verify that M has the right shape.
    if (M.size() != NumRows)
      CoCoA_ERROR(ERR::BadRowIndex, "NewDenseMatrix(R,n,m,M)");
    for (size_t i=0; i < NumRows; ++i)
      if (M[i].size() != NumCols)
        CoCoA_ERROR(ERR::BadColIndex, "NewDenseMatrix(R,n,m,M)");
    matrix ans(new DenseMatrixImpl(R, NumRows, NumCols));
    for (size_t i=0; i < NumRows; ++i)
      for (size_t col=0; col < NumCols; ++col)
        SetEntry(ans, i, col, M[i][col]);
    return ans;
  }


  // I bet this makes lots of wasteful copies...
  matrix NewDenseMatrix(ConstMatrix M)
  {
    const size_t r = NumRows(M);
    const size_t c = NumCols(M);
    matrix ans(new DenseMatrixImpl(BaseRing(M), r, c));
    for (size_t i=0; i < r; ++i)
      for (size_t j=0; j < c; ++j)
        SetEntry(ans, i, j, M(i, j));
    return ans;
  }

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/DenseMatrix.C,v 1.1.1.1 2007/03/09 15:16:11 abbott Exp $
// $Log: DenseMatrix.C,v $
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.6  2007/03/08 18:22:30  cocoa
// Just whitespace cleaning.
//
// Revision 1.5  2007/02/10 18:44:03  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.4  2006/11/27 16:18:33  cocoa
// -- moved classes declarations from .H to .C (DenseMatrix, DiagMatrix,
//    FieldIdeal, SpecialMatrix)
//
// Revision 1.3  2006/10/16 23:18:59  cocoa
// Corrected use of std::swap and various special swap functions.
// Improved myApply memfn for homs of RingDistrMPolyInlPP.
//
// Revision 1.2  2006/10/06 14:04:15  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.2  2006/01/25 15:58:19  cocoa
// -- fixed: size condition in mul(matrix&, ConstMatrix, ConstMatrix), by Kaspar
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.2  2005/05/06 11:57:15  cocoa
// Fixed bug in ctor for DenseMatrixImpl.
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.9  2005/04/20 15:40:48  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.8  2005/04/19 15:39:55  cocoa
// Matrices now use reference counts.
//
// Revision 1.7  2005/04/19 14:06:04  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.6  2005/03/29 17:26:29  cocoa
// -- changed: myDet, myRank call MatrixArith functions
//
// Revision 1.5  2005/03/11 16:44:18  cocoa
// New abstract class structure for matrices.
// New types of special matrix.
//
// Revision 1.4  2005/03/02 18:46:41  cocoa
// Added new types ConstRefMatrix, and RefMatrix following along
// the lines of ConstRefRingElem and RefRingElem.  The semantics
// should be a bit clearer now.
//
// Revision 1.3  2005/02/11 14:15:20  cocoa
// New style ring elements and references to ring elements;
// I hope I have finally got it right!
//
// Revision 1.2  2005/02/09 11:19:30  cocoa
// -- fixed: determinant over a GCDRing (through FrF)
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.6  2004/11/29 16:22:35  cocoa
// -- added function for computing adjoint and inverse for DenseMatrix
//    (so adjoint/inverse matrix is computed by OrdvArith and is no
//    longer needed by PPOrdering)
//
// Revision 1.5  2004/11/02 14:49:03  cocoa
// -- new code for matrix orderings
//
// Revision 1.4  2004/07/20 15:04:06  cocoa
// The next step in the new "ring element" conversion process:
// handling the case of creating a "const RefRingElem" object
// (since C++ refuses to do this properly itself).
//
// Revision 1.3  2004/03/20 17:46:11  cocoa
// Check in prior to departure to RWCA
//
// Revision 1.2  2004/03/09 16:15:41  cocoa
// First version of matrices.  A simple exmaple compiles and runs,
// so I'm checking in.
//
// Revision 1.1  2004/01/28 16:29:17  cocoa
// First attempt at implementing matrices.
//
//
