//   Copyright (c)  1997-2006  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License (version 2)
//   as published by the Free Software Foundation.  A copy of the full
//   licence may be found in the file COPYING in this directory.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoA; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include <stdio.h>
#include <math.h>
#include "DUPZfactor_lift.h"
#include "DUPIsubtract_product.h"
#include "jalloc.h"
#include "jaaerror.h"
#include "jtime.h"
#include "DUPI_DUPFF.h"
#include "DUPZfactor_info.h"
#include "DUPZpadic_expand.h"



/***************************************************************************/
/* Conversion of poly from a y-adic expansion to a (y-alpha)-adic expansion*/
/* Input is poly  = F[0] +        y *F[1] + ... +         y^deg * F[deg].  */
/* On return poly = F[0] + (y-alpha)*F[1] + ... + (y-alpha)^deg * F[deg].  */


void DUPFFbiv_shift(const DUPFF *F, int deg, uint32 alpha)
{
  int i, j;

  for (i=0; i < deg; i++)
    for (j=deg-1; j >= i; j--)
      DUPFFshift_add(F[j],F[j+1],0,alpha);/* F[j] += alpha * F[j+1] */
}

/***************************************************************************/
/* Lift a univariate factorization to a bivariate one.                     */
/* Input is f a y-adic expansion of the bivariate polynomial,              */
/* and the univariate factorization of f at y=alpha.                       */
/* Output is array such that lifted_factors[i] is y-adic expn of lifted factor*/

DUPFF** DUPFFbivariate_lift(DUPFF *f, int dy, DUPFFlist factors, uint32 alpha)
{
  DUPFF **lifted_factors;
  DUPFFbiv_lifter lifter;
  int i;

  if (alpha != 0)
    DUPFFbiv_shift(f, dy, alpha); /* convert to (y-alpha)-adic expansion */
  lifter = DUPFFbiv_lift_init(f, factors, dy);
  for (i=1; i < dy; i++)
    DUPFFbiv_lift_step(lifter);

  DUPFFbiv_lift_output(lifted_factors, lifter);
  DUPFFbiv_lifter_dtor(lifter);
  if (alpha != 0)
    for (i=0; i < nfactors; i++)
      DUPFFbiv_shift(lifted_factors[i], p-alpha);
  return lifted_factors;
}



/***************************************************************************/

void solve_for_correction(DUPFF *gdelta, const DUPFF E, const DUPFF g0, const DUPFF grecip)
{
  DUPFF junk, tmp;
  int dg0, dE;

  dg0 = DUPFFdeg(g0);
  dE = DUPFFdeg(E);
  junk = DUPFFnew(dE + dg0);  /* these are all a bit too big */
  tmp = DUPFFnew(2*dg0);      /*                             */
  *gdelta = DUPFFnew(dg0);    /*                             */

  DUPFFdiv4(junk, tmp, E, g0);
  DUPFFmul3(tmp, tmp, grecip);
  DUPFFdiv4(junk, *gdelta, tmp, g0);

  DUPFFfree(tmp);
  DUPFFfree(junk);
}

/***************************************************************************/


DUPFFbiv_lifter DUPFFbiv_lifter_ctor(DUPFF *g, DUPFF *h, DUPFFbiv_lifter g_lifter, DUPFFbiv_lifter h_lifter, int rmax)
{
  DUPFFfactor_lifter ans;
  DUPFF grecip, hrecip, tmp;
  int df;
  
  ans = (DUPFFbiv_lifter)MALLOC(sizeof(struct DUPFFbiv_lift_struct));
  ans->r = 1;
  ans->g = (DUPFF*)MALLOC(rmax*sizeof(DUPFF));
  if (g_lifter) ans->g[0] = DUPFFmul(g_lifter->g[0], g_lifter->h[0]);
  else ans->g[0] = DUPFFcopy(*g);
  ans->h = (DUPFF*)MALLOC(rmax*sizeof(DUPFF));
  if (h_lifter) ans->h[0] = DUPFFmul(h_lifter->g[0], h_lifter->h[0]);
  else ans->h[0] = DUPFFcopy(*h);
  df = DUPFFdeg(ans->g[0]) + DUPFFdeg(ans->h[0]);
  tmp = DUPFFexgcd(&grecip, &hrecip, ans->g[0], ans->h[0]);
  if (DUPFFdeg(tmp) != 0) { JERROR(JERROR_HENSEL); exit(1); }
  /* since the gcd is not monic we have to do this... */
  DUPFFdiv2ff(grecip, DUPFFlc(tmp));
  DUPFFdiv2ff(hrecip, DUPFFlc(tmp));
  DUPFFfree(tmp);
  ans->grecip = grecip;
  ans->hrecip = hrecip;
  ans->g_lifter = g_lifter;
  ans->h_lifter = h_lifter;
  ans->E = DUPFFnew(df);

  return ans;
}


void DUPFFbiv_lift_dtor(DUPFFbiv_lifter this)
{
  int i;
  int r;

  if (this == NULL) return;
  r = this->r;
  DUPFFfree(this->E);
  if (this->g_lifter) DUPFFbiv_lift_dtor(this->g_lifter);
  if (this->h_lifter) DUPFFbiv_lift_dtor(this->h_lifter);
  for (i=0; i < r; i++)
  {
    DUPFFfree(this->g[i]);
    DUPFFfree(this->h[i]);
  }
  FREE(this->g);
  FREE(this->h);
  DUPFFfree(this->grecip);
  DUPFFfree(this->hrecip);
  FREE(this);
}

/***************************************************************************/

void DUPFFbiv_lift_step(DUPFFbiv_lifter this, DUPFF f_cmpt_r)
{
  int i;
  /* the rest are just aliases to make the code easier to read (!?!) */
  DUPFF       *g     = this->g;
  DUPFF       *h     = this->h;
  int         r      = this->r;
  const DUPFF grecip = this->grecip;
  const DUPFF hrecip = this->hrecip;
  DUPFF       E      = this->E;


  DUPFFcopy2(E, f_cmpt_r);
  for (i=1; i < r; i++)
    DUPFFsubtract_product(E, g[i], h[r-i]);

  solve_for_correction(&g[r], E, g[0], hrecip);
  solve_for_correction(&h[r], E, h[0], grecip);
  this->r++;

  if (this->g_lifter) DUPFFbiv_lift_step(this->g_lifter, g[r]);
  if (this->h_lifter) DUPFFbiv_lift_step(this->h_lifter, h[r]);
}


/***************************************************************************/
/* Convert the p-adic expansions of the lifted factors to their normal     */
/* representations.                                                        */

int DUPFFbiv_lift_output(DUPFF *factors, const DUPFFbiv_lifter this, int i)
{
  int p, k;

  p = CurrentFF.prime;
  k = this->r;
  if (this->g_lifter) i = DUPFFbiv_lift_output(factors, this->g_lifter, i);
  else factors[i++] = DUPFF_from_padic(this->g, p, k);
  if (this->h_lifter) i = DUPFFbiv_lift_output(factors, this->h_lifter, i);
  else factors[i++] = DUPFF_from_padic(this->h, p, k);
  return i;
}



/***************************************************************************/

static DUPFFbiv_lifter DUPFFbiv_lift_init1(const DUPFF *f, const DUPFFlist pfactors, int rmax)
{
  DUPFFbiv_lifter *lifter, ans, new;
  DUPFF **pfac;
  int i, j, nfactors, *deg, newdeg;
  DUPFFlist iter;

  nfactors = DUPFFlist_length(pfactors);
  pfac = (DUPFF**)MALLOC(nfactors*sizeof(DUPFF*));
  deg = (int*)MALLOC(nfactors*sizeof(int));
  for (i=0, iter=pfactors; iter; iter = iter->next, i++)
  {
    pfac[i] = &iter->poly;
    deg[i] = DUPFFdeg(iter->poly);
  }

  lifter = (DUPFFbiv_lifter*)MALLOC(nfactors*sizeof(DUPFFbiv_lifter));
  for (i=0; i < nfactors; i++) lifter[i] = NULL;
  for (i=0; i < nfactors-1; i++)
  {
    new = DUPFFbiv_lifter_ctor(pfac[i], pfac[i+1], lifter[i], lifter[i+1], rmax);
    newdeg = deg[i] + deg[i+1];
    /* Re-sort the array; i.e. insert new lifter in correct position */
    for (j=i+2; j < nfactors; j++)
    {
      if (newdeg <= deg[j]) break;
      deg[j-1] = deg[j];
      pfac[j-1] = pfac[j];
      lifter[j-1] = lifter[j];
    }
    lifter[j-1] = new;
    deg[j-1] = newdeg;
  }
  ans = lifter[nfactors-1];

  FREE(lifter); /* do not free its elements! */
  FREE(pfac);
  FREE(deg);

  return ans;
}


void DUPFFbiv_lift_init(DUPFF *f, int dy, DUPFF *factors, uint32 alpha)
{
  const int p = CurrentFF.prime;

  /* We are sure to get a non-trivial factorization before height rmax: */
  this->lifter = DUPFFbiv_lift_init1(this->f, this->pfactors, this->rmax);
  this->f_expansion = DUPFFpadic_expand_init(this->f, this->p, this->rmax);
#ifdef FACTOR_TIMER
  printf("Lift init done: \t"); print_time();
#endif
}




/****************************************************************************/
/* This function is called after a speculative early search which has found */
/* at least one factor but further lifting is still required.               */
/* It creates a new lifting tree for the remaining factors.                 */

void DUPFFbiv_lift_revise1(DUPFFfactor_info info, int height)
{
  int i;

  DUPFFbiv_lift_init(info);
  DUPFFpadic_expand_end(info->f_expansion);
  info->f_expansion = DUPFFpadic_expand_init(info->f, info->p, info->rmax);
  DUPFFpadic_expand_step(info->f_expansion);
  if (height > info->rmax) height = info->rmax;
  for(i=1; i < height; i++)
  {
    DUPFFpadic_expand_step(info->f_expansion);
    DUPFFbiv_lift_step(info->lifter, info->f_expansion->monic_cmpts[i]);
  }
}


void DUPFFbiv_lift_revise(DUPFFfactor_info info)
{
  int height = info->lifter->r;

  DUPFFbiv_lift_dtor(info->lifter); info->lifter = NULL;
  DUPFFpadic_expand_end(info->f_expansion); info->f_expansion = NULL;
  DUPFFbiv_lift_revise1(info, height);
}


/***************************************************************************/
void liftprint(DUPFFbiv_lifter this, int indent)
{
int i;
if (this == NULL) return;
putchar('>');for(i=0;i<indent;i++)putchar(' ');
for(i=0;i<this->r;i++)printf("%d,",DUPFFdeg(this->g[i]));printf("\n");
for(i=0;i<=indent;i++)putchar(' ');
for(i=0;i<this->r;i++)printf("%d,",DUPFFdeg(this->h[i]));printf("\n");
liftprint(this->g_lifter, indent+2);
liftprint(this->h_lifter, indent+2);
}


/***************************************************************************/
/***************************************************************************/
static DUPFFbiv_lifter DUPFFbiv_lift_init1(const DUPFF *f, const DUPFFlist pfactors, int rmax)
{
  DUPFFbiv_lifter *lifter, ans, new;
  DUPFF **pfac;
  int i, j, nfactors, *deg, newdeg;
  DUPFFlist iter;
#ifdef JAA_FLAG
  DUPFF fi, fj, tmp, *xyz;
  int k, t, tmin, jmin, d;
#endif

  nfactors = DUPFFlist_length(pfactors);
  pfac = (DUPFF**)MALLOC(nfactors*sizeof(DUPFF*));
  deg = (int*)MALLOC(nfactors*sizeof(int));
  for (i=0, iter=pfactors; iter; iter = iter->next, i++)
  {
    pfac[i] = &iter->poly;
    deg[i] = DUPFFdeg(iter->poly);
  }

  lifter = (DUPFFbiv_lifter*)MALLOC(nfactors*sizeof(DUPFFbiv_lifter));
  for (i=0; i < nfactors; i++) lifter[i] = NULL;
  for (i=0; i < nfactors-1; i++)
  {
#ifdef JAA_FLAG
    if (i+1 == nfactors-1) goto skip;;
    d = deg[i+1];
    tmin = 0;  jmin = -987654321;
    for (j=i+1; j < nfactors; j++)
    {
      if (deg[j] > d) break;
      if (lifter[i]) fi = DUPFFmul(lifter[i]->g[0], lifter[i]->h[0]);
      else fi = pfac[i][0];
      if (lifter[j]) fj = DUPFFmul(lifter[j]->g[0], lifter[j]->h[0]);
      else fj = pfac[j][0];
      tmp = DUPFFmul(fi, fj);
      t = 0;
      for (k=0; k <= DUPFFdeg(tmp); k++) if (tmp->coeffs[k]) t++;
//printf("t is %d\n", t);
      if (tmin != 0 && t >= tmin) continue;
      tmin = t;
      jmin = j;
    }
//printf("chosen t=%d\n", tmin);
    new = lifter[jmin]; newdeg = deg[jmin];
    xyz = pfac[jmin];
    for (j=jmin; j>i+1; j--)
    {
      deg[j] = deg[j-1];
      pfac[j] = pfac[j-1];
      lifter[j] = lifter[j-1];
    }
    deg[i+1] = newdeg;
    lifter[i+1] = new;
    pfac[i+1] = xyz;
  skip:
#endif
    new = DUPFFbiv_lifter_ctor(pfac[i], pfac[i+1], lifter[i], lifter[i+1], rmax);
    newdeg = deg[i] + deg[i+1];
#if 0
lifter[i+1]=new;deg[i+1]=newdeg;/* DELETE ME -- BUG BUG BUG BUG BUG */
#else
    /* Re-sort the array; i.e. insert new lifter in correct position */
    for (j=i+2; j < nfactors; j++)
    {
      if (newdeg <= deg[j]) break;
      deg[j-1] = deg[j];
      pfac[j-1] = pfac[j];
      lifter[j-1] = lifter[j];
    }
    lifter[j-1] = new;
    deg[j-1] = newdeg;
#endif
  }
  ans = lifter[nfactors-1];

  FREE(lifter); /* do not free its elements! */
  FREE(pfac);
  FREE(deg);

  return ans;
}

/***************************************************************************/
/***************************************************************************/

  f_cmpt = (DUPFF*)MALLOC(rmax*sizeof(DUPFF));/* BUG how to initialize? */
  fac = (DUPFF*)MALLOC(nfacs*sizeof(DUPFF));/* BUG how to initialize? */
  p = nextprime(DUPFFdeg(f[0]));
  for (;1; p = nextprime(p))
  {
    FFp = FFctor(p);
    FFselect(FFp);
    fp = DUPFF_to_DUPFF(f[0]);
    if (mpz_fdiv_ui(DUPFFlc(f[0]), p) == 0 || !DUPFFsqfr(fp))
    {
      FFdtor(FFp);
      continue;
    }
    for (i=0; i < rmax; i++)
      f_cmpt[i] = DUPFF_to_DUPFF(f[i]);
    for (i=0; i < nfacs; i++)
      fac[i] = DUPFF_to_DUPFF(factors[i]);
    lifter = lift_init(f_cmpt, rmax, fac, nfacs);
    for (i=1; i < rmax; i++)
      lift_step(lifter, f_cmpt[i]);
    
    for (i=0; i < nfacs; i++)
      for (j=0; j < rmax; j++)
        DUPFFcra(factor[i][j], modulus, fac[i][j], p);
    
  }
