#include "CoCoA/BuildInfo.H"
#include "CoCoA/FractionField.H"
#include "CoCoA/GlobalManager.H"
#include "CoCoA/RingDistrMPoly.H"
#include "CoCoA/RingQ.H"
#include "CoCoA/SparsePolyRing.H"
#include "CoCoA/TmpHilbert.H"
#include "CoCoA/io.H"
#include "CoCoA/symbol.H"

using namespace CoCoA;

#include <algorithm>
using std::min;
#include <iostream>
using std::endl;
#include <vector>
using std::vector;

//----------------------------------------------------------------------
// First test for Hilbert.
//----------------------------------------------------------------------

#define TEST_ASSERT(cond)   do { if (!(cond)) CoCoA::AssertionFailed(#cond, __FILE__, __LINE__); } while(0)

// ---- chess tests -----------------------------------------------------

ConstRefRingElem CsbSquareIndet(SparsePolyRing P, size_t l, size_t sq1, size_t sq2)
{
  CoCoA_ASSERT( l*l <= NumIndets(P) );
  CoCoA_ASSERT( sq1 <= l && sq2 <= l );
  return indet(P, (sq1-1)*l + (sq2-1));
}


ideal NewQueenMovesFrom(SparsePolyRing P, size_t Csb, size_t sq1, size_t sq2)
{
  ConstRefRingElem x = CsbSquareIndet(P, Csb, sq1, sq2);
  vector<RingElem> g;
  for ( size_t i=sq2+1 ; i<=Csb ; ++i )
    g.push_back(x * CsbSquareIndet(P, Csb, sq1, i));
  for ( size_t i=sq1+1 ; i<=Csb ; ++i )
    g.push_back(x * CsbSquareIndet(P, Csb, i, sq2));
  for ( size_t i=min(Csb-sq1,Csb-sq2) ; i>0 ; --i )
    g.push_back(x * CsbSquareIndet(P, Csb, sq1+i, sq2+i));
  for ( size_t i=min(Csb-sq1, sq2-1) ; i>0 ; --i )
    g.push_back(x * CsbSquareIndet(P, Csb, sq1+i, sq2-i));
  return ideal(P, g);
}


ideal NewQueenIdeal(SparsePolyRing P, size_t Csb)
{
  ideal I = ideal(P);
  for ( size_t sq1=1 ; sq1<=Csb ; ++sq1 )
    for ( size_t sq2=1 ; sq2<=Csb ; ++sq2 )
      I += NewQueenMovesFrom(P, Csb, sq1, sq2);
  return I;
}


// ----------------------------------------------------------------------

void program()
{
  GlobalManager CoCoAFoundations;

  ring Q = RingQ();
  // the following 3 lines are to get Q[lambda].
  // It should get easier...
  SparsePolyRing HPRing = NewPolyRing(Q, symbols("lambda"));
  RingElem lambda = indet(HPRing, 0);

  SparsePolyRing P = NewPolyRing(Q, 4);
  const vector<RingElem>& x = indets(P);
  vector<RingElem> g;
  g.push_back(x[1]);
  g.push_back(x[2]);
  g.push_back(x[3]);
  ideal I = ideal(P, g);
  TEST_ASSERT(HilbertNumeratorMod(HPRing, I) == -power(lambda,3) +3*power(lambda,2) -3*lambda +1);
  
  SparsePolyRing CsbRing = NewPolyRing(Q, 36);
  ideal Q6 = NewQueenIdeal(CsbRing, 6);
  RingElem HN = HilbertNumeratorMod(HPRing, Q6);
  TEST_ASSERT(LC(HN) == 19);
  TEST_ASSERT(deg(HN) == 36);
}


//----------------------------------------------------------------------
// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home/cocoa/cvs-repository/CoCoALib-0.99/src/tests/test-hilbert1.C,v 1.1.1.1 2007/03/09 15:16:12 abbott Exp $
// $Log: test-hilbert1.C,v $
// Revision 1.1.1.1  2007/03/09 15:16:12  abbott
// Imported files
//
// Revision 1.9  2007/03/08 17:41:36  bigatti
// -- improved NewPolyRing
//
// Revision 1.8  2007/03/08 14:38:07  cocoa
// Added new range function in symbol.H, and tidied many calls to PolyRing
// pseudo ctors (as a consequence).
//
// Revision 1.7  2007/03/03 14:13:21  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.6  2007/03/02 17:46:40  bigatti
// -- unique RingZ and RingQ
// -- requires foundations.H ;  foundations blah;  (thik of a better name)
//
// Revision 1.5  2007/02/26 17:11:58  bigatti
// -- getting ready for unique ring Z: using NewZmod(N), NewRingQ()
//
// Revision 1.4  2007/02/10 18:44:02  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.3  2007/01/17 17:38:11  bigatti
// -- moved all cocoa-4 code for hilbert into src/TmpHilbertDir
//
// Revision 1.2  2006/12/07 17:25:32  cocoa
// -- minimal set of #include's instead of library.H
//
// Revision 1.1  2006/11/16 18:16:10  cocoa
// -- added test for HilbertNumeratorMod
//
