#include "CoCoA/BuildInfo.H"
#include "CoCoA/DenseMatrix.H"
#include "CoCoA/GlobalManager.H"
#include "CoCoA/RingFp.H"
#include "CoCoA/RingZ.H"
#include "CoCoA/SpecialMatrix.H"
#include "CoCoA/io.H"
#include "CoCoA/matrix.H"

using namespace CoCoA;
#include <iostream>
using std::endl;
#include <cstddef>
using std::size_t;
#include <vector>
using std::vector;

#define TEST_ASSERT(cond)   do { if (!(cond)) CoCoA::AssertionFailed(#cond, __FILE__, __LINE__); } while(0)


void TestBlockMatrix()
{
  const ring Z = RingZ();
  matrix M(NewDenseMatrix(Z, 6, 6));
  for (size_t i=0; i < 6; ++i)
    for (size_t j=0; j < 6; ++j)
      SetEntry(M, i, j, 6*i+j);

  vector<size_t> R1, R2, R3;
  R1.push_back(0);
  R2.push_back(1);
  R2.push_back(2);
  R3.push_back(3);
  R3.push_back(4);
  R3.push_back(5);
  vector<size_t> C1 = R1;
  vector<size_t> C2 = R2;
  vector<size_t> C3 = R3;

  ConstMatrix M11(NewSubmatrix(M, R1, C1));
  ConstMatrix M12(NewSubmatrix(M, R1, C2));
  ConstMatrix M13(NewSubmatrix(M, R1, C3));
  ConstMatrix M21(NewSubmatrix(M, R2, C1));
  ConstMatrix M22(NewSubmatrix(M, R2, C2));
  ConstMatrix M23(NewSubmatrix(M, R2, C3));
  ConstMatrix M31(NewSubmatrix(M, R3, C1));
  ConstMatrix M32(NewSubmatrix(M, R3, C2));
  ConstMatrix M33(NewSubmatrix(M, R3, C3));

  ConstMatrix Mcopy = NewBlockMatrix(NewBlockMatrix(M11, M12, M21, M22),
                                     NewBlockMatrix(M13, NewZeroMatrix(Z,1,0), M23, NewZeroMatrix(Z,2,0)),
                                     NewBlockMatrix(M31, M32, NewZeroMatrix(Z,0,1), NewZeroMatrix(Z,0,2)),
                                     M33);

  ConstMatrix Mcopy2 = NewBlockMatrix(M11,
                                      NewBlockMatrix(M12, M13, NewZeroMatrix(Z,0,2), NewZeroMatrix(Z,0,3)),
                                      NewBlockMatrix(M21, NewZeroMatrix(Z,2,0), M31, NewZeroMatrix(Z,3,0)),
                                      NewBlockMatrix(M22, M23, M32, M33));

  for (size_t i=0; i < 6; ++i)
    for (size_t j=0; j < 6; ++j)
    {
      TEST_ASSERT(M(i, j) == Mcopy(i, j));
      TEST_ASSERT(M(i, j) == Mcopy2(i, j));
    }
}


void program()
{
  GlobalManager CoCoAFoundations;

  const ring Z = RingZ();
  TestBlockMatrix();

  matrix M(NewDenseMatrix(Z, 2, 3));
  SetEntry(M,1,0,3);
  SetEntry(M,1,1,1);
  SetEntry(M,1,2,4);

  TEST_ASSERT(IsZero(M(0,0)));

  TEST_ASSERT(M->myIsZeroRow(0));

  M->myAddRowMul(0, 1, RingElem(Z, 2));
  TEST_ASSERT(!M->myIsZeroRow(0));
  TEST_ASSERT(M(0,1) == 2);
  M->myAddRowMul(0, 1, M(0,0)); // alias test
  TEST_ASSERT(M(0,0) == 8*M(1,0));
  TEST_ASSERT(M(0,1) == 8*M(1,1));
  TEST_ASSERT(M(0,2) == 8*M(1,2));

  // An IdentityMatrix is constant; you cannot change it, but you can copy it.
  ConstMatrix I10 = NewIdentityMatrix(NewZmod(7), 10);
  // SetEntry(I10,1,1,1); <-- as expected, this does not compile because I10 is constant
  for (size_t i=0; i < 10; ++i)
    for (size_t j=0; j < 10; ++j)
    {
      if (i == j) TEST_ASSERT(IsOne(I10(i, j)));
      else TEST_ASSERT(IsZero(I10(i, j)));
    }

  // A few quick checks on a zero matrix
  ConstMatrix ZM = NewZeroMatrix(Z, 12345,67890);
  for (size_t i=0; i < 12345; ++i)
    TEST_ASSERT(IsZeroRow(ZM, i));
  for (size_t j=0; j < 67890; ++j)
    TEST_ASSERT(IsZeroCol(ZM, j));
  for (size_t i=0; i < 12345; ++i)
    TEST_ASSERT(IsZero(ZM(i, i)));

  // Create a submatrix view of I10; since the row and column
  // indices are the same lists, the result will be another
  // identity matrix.
  vector<size_t> RowIndices;
  RowIndices.push_back(2);
  RowIndices.push_back(7);
  RowIndices.push_back(1);
  RowIndices.push_back(8);
  vector<size_t> ColIndices = RowIndices;
  ConstMatrix SubM = NewSubmatrix(I10, RowIndices, ColIndices);
  for (size_t i=0; i < NumRows(SubM); ++i)
    for (size_t j=0; j < NumCols(SubM); ++j)
    {
      if (i == j) TEST_ASSERT(IsOne(SubM(i, j)));
      else TEST_ASSERT(IsZero(SubM(i, j)));
    }

  // Some simple tests on the transposed view of a matrix.
  ConstMatrix TrM = NewTransposeMatrix(M);
  TEST_ASSERT(NumRows(TrM) == NumCols(M));
  TEST_ASSERT(NumRows(M) == NumCols(TrM));
  ConstMatrix TrTrM = NewTransposeMatrix(TrM);
  for (size_t i = 0; i < NumRows(M); ++i)
    for (size_t j =0 ; j < NumCols(M); ++j)
      TEST_ASSERT(TrTrM(i, j) == M(i, j));

  TestBlockMatrix();
}


// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA Error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }

  BuildInfo::PrintAll(GlobalErrput());
  return 1;
}
